"""
Example sentences to test spaCy and its language models.

>>> from spacy.lang.ti.examples import sentences
>>> docs = nlp.pipe(sentences)
"""


sentences = [
    "አፕል ብዩኬ ትርከብ ንግድ ብ1 ቢሊዮን ዶላር ንምግዛዕ ሐሲባ።",
    "ፈላማይ ክታበት ኮቪድ 19 ተጀሚሩ፤ሓዱሽ ተስፋ ሂቡ ኣሎ",
    "ቻንስለር ጀርመን ኣንገላ መርከል ዝርግሓ ቫይረስ ኮሮና ንምክልካል ጽኑዕ እገዳ ክግበር ጸዊዓ",
    "ለንደን ብዓዲ እንግሊዝ ትርከብ ዓባይ ከተማ እያ።",
    "ናበይ አለኻ፧",
    "ናይ ፈረንሳይ ፕሬዝዳንት መን እዩ፧",
    "ናይ አሜሪካ ዋና ከተማ እንታይ እያ፧",
    "ኦባማ መዓስ ተወሊዱ፧",
]
